!----------------------------------------------------------------------------------------------------------
                                             program bluntbody 
!--
!--                                             Version 2.0
!--                            Copyright (c) 2009-2016 Michael M. Schulreich
!--
!-----------------------------------------------------------------------------------------------------------
!--
!--  Solver of the hypersonic blunt body problem based on the method first described in Schneider, W. 1968, 
!--  J. Fluid Mech., 31, 397 (Erratum: Schneider, W. 1968, J. Fluid Mech., 32, 829). 
!--
!--  For astrophysical application see:
!--  > Schulreich, M. M. & Breitschwerdt, D. 2011, Astron. Astrophys., 531, 13
!--
!--  Check the README for information about input and output variables.
!--
!-----------------------------------------------------------------------------------------------------------

  implicit none

  real(kind=8),parameter :: meshsize   = 1.d-3         ! lowering this can increase CPU time dramatically
  logical,parameter      :: verbose    = .true.        ! switch for output of additional program info
                                                       ! may be turned off when running program on a cluster

  real(kind=8),parameter :: mH         = 1.6737236d-24 ! hydrogen mass
  real(kind=8),parameter :: kB         = 1.3806485d-16 ! Boltzmann constant

  ! Better not change the values below
  character(len=300)     :: fname      = "input.dat"   ! name of optional input file
  real(kind=8),parameter :: floor_hat  = 1.d-3         ! floor values to avoid singularities
  real(kind=8),parameter :: floor_star = 1.d-20
  real(kind=8),parameter :: tol        = 1.d-10        ! tolerance for Romberg integrator
  
  integer(kind=4)          :: j,plines,iter,i,t1,t2,clock_rate,clock_max
  real(kind=8)             :: gam,M,uinf,rhoinf,mu,a,alpha,scale_l,zhatmin,zhatmax
  real(kind=8),allocatable :: zhatstep(:)
  real(kind=8)             :: const,rhat,zc,rc,sinbhat2,cosbhat2,chihat,Phat,rhohat,uhat,That,sspeed
  real(kind=8)             :: rstar,kappahat,llim,ulim,integr1sol,P,Pb,rhob,ub,Tb,integr2sol,delta
  real(kind=8)             :: zhat,rho,u,T,y,romberg_trap
  character(len=5)         :: out
  character(len=300)       :: cwd
  character                :: choice
  logical                  :: iexist,ileave


  if(verbose) then  
     write(*,'(a)') "---------------------------------------------------------------------------"
     write(*,'(a)') "  Solver for the astrophysical hypersonic blunt body problem based on the  "
     write(*,'(a)') "   method first described in Schneider, W. 1968, J. Fluid Mech., 31, 397   "
     write(*,'(a)') "          (Erratum: Schneider, W. 1968, J. Fluid Mech., 32, 829)           "
     write(*,'(a)') "                                                                           "
     write(*,'(a)') "                                Version 2.0                                "
     write(*,'(a)') "               Copyright (c) 2009-2016 Michael M. Schulreich               "
     write(*,'(a)') "---------------------------------------------------------------------------"
     write(*,'(a)') ""
  end if

  
  ! First check if input file is available ...
  call getcwd(cwd)
  fname=trim(cwd)//"/"//trim(fname)
  inquire(file=fname, exist=iexist)

  select case (iexist)

     case (.false.) 

        write(*,'(a)') "Enter FLOW GEOMETRY PARAMETER;"  
        ileave=.false.
        do while(.not.ileave)
           write(*,'(a)') "choose either j = 0 (plane flow) or 1 (axisymmmetric flow);"
           write(*,'(a)',advance="no") "j = "
           read(*,*) j
           if (j == 0 .or. j == 1) ileave=.true.
        end do        
        write(*,'(a)') ""

        write(*,'(a)') "Enter ADIABATIC INDEX;"
        write(*,'(a)',advance="no") "gam = "
        read(*,*) gam 
        write(*,'(a)') ""

        write(*,'(a)') "Enter UPSTREAM MACH NUMBER;"
        write(*,'(a)',advance="no") "M = "
        read(*,*) M
        write(*,'(a)') ""

        write(*,'(a)') "Enter UPSTREAM FLOW (or OBSTACLE) VELOCITY (in cm s-1);"
        write(*,'(a)',advance="no") "uinf = "
        read(*,*) uinf
        write(*,'(a)') ""

        write(*,'(a)') "Enter UPSTREAM FLUID DENSITY (in g cm-3);"
        write(*,'(a)',advance="no") "rhoinf = "
        read(*,*) rhoinf
        write(*,'(a)') ""

        ileave=.false.
        do while(.not.ileave)
           write(*,'(a)',advance="no") "METALLICITY of the fluid is solar (y/n)? "
           read(*,*) choice
           select case (choice)
              case("y")
                 mu = 0.62d0
                 ileave = .true.
              case("n")
                 write(*,'(a)') "Enter MEAN MOLECULAR WEIGHT of the fluid;"
                 write(*,'(a)',advance="no") "mu = " 
                 read(*,*) mu
                 ileave=.true.
           end select
        end do
        write(*,'(a)') ""

        write(*,'(a)') "Enter BOW SHOCK GEOMETRY PARAMETERS;"
        write(*,'(a)') "parabola equation: r=(a*z)**(1/alpha);"
        write(*,'(a)',advance="no") "a = "
        read(*,*) a
        write(*,'(a)',advance="no") "alpha = "
        read(*,*) alpha
        write(*,'(a)') ""

        write(*,'(a)') "Enter LENGTH SCALE of the problem (in cm);"
        write(*,'(a)') "enter 3.08567758d18 for LENGTH SCALE in pc"
        write(*,'(a)',advance="no") "scale_l = "
        read(*,*) scale_l
        write(*,'(a)') ""

        write(*,'(a)') "Enter LOWER LIMIT for SHOCK-NORMAL computations (in units of scale_l);"
        write(*,'(a)',advance="no") "zhatmin = "
        read(*,*) zhatmin
        write(*,'(a)') ""

        write(*,'(a)') "Enter UPPER LIMIT for SHOCK-NORMAL computations (in units of scale_l);"
        write(*,'(a)',advance="no") "zhatmax = "
        read(*,*) zhatmax
        write(*,'(a)') ""
  
        write(*,'(a)') "Enter desired AMOUNT of SHOCK NORMALS (must not exceed 99999);"
        write(*,'(a)',advance="no") "plines = "
        read(*,*) plines
        write(*,'(a)') ""

        ! Save input data in a corresponding file
        open(10,file=fname,action="write",status="replace")
        write(10,*) j
        write(10,*) gam
        write(10,*) M
        write(10,*) uinf
        write(10,*) rhoinf
        write(10,*) mu
        write(10,*) a
        write(10,*) alpha
        write(10,*) scale_l
        write(10,*) zhatmin
        write(10,*) zhatmax
        write(10,*) plines
        close(10)
        
     case(.true.)

        open(10,file=fname,action="read")
        read(10,*) j
        read(10,*) gam
        read(10,*) M
        read(10,*) uinf
        read(10,*) rhoinf
        read(10,*) mu
        read(10,*) a
        read(10,*) alpha
        read(10,*) scale_l
        read(10,*) zhatmin
        read(10,*) zhatmax
        read(10,*) plines
        close(10)

  end select

  const = mu*mH/kB

  if(verbose) write(*,'(a)') ">> Starting calculations."
  call system_clock (t1,clock_rate,clock_max)
  
! Initialize logarithmic spacing of probe lines by introducing a temporary equivalent interval [1:10]
! that is used to calculate percentages of the size of the actual interval, using the fact that
! log10(1)=0 and log10(10)=1.

  allocate(zhatstep(plines))
  do i = 1,plines
     zhatstep(i) = (10.d0-1.d0)/(real(plines-1,kind=8))
     zhatstep(i) = log10(1.d0+real(i-1,kind=8)*zhatstep(i))
  end do


! (1) Write out profiles for fluid variables immediately behind the bow shock wave
! ================================================================================

  open(11,file="shock.dat",status="replace",action="write")
  
  rhat = 0.d0
  
  do
!    generate bow shock
     zc       = z(rhat)
     if(zc > zhatmax) exit
     rc       = rhat
     
     sinbhat2 = sinb2(rhat)
     cosbhat2 = cosb2(rhat)
     chihat   = (gam-1.d0)/(gam+1.d0)+2.d0/((gam+1.d0)*M**2*sinbhat2)
     Phat     = 1.d0/(gam*M**2)+(1.d0-chihat)*sinbhat2
     rhohat   = 1.d0/chihat
     uhat     = sqrt(cosbhat2)
     That     = Phat/rhohat
     sspeed   = sqrt(gam*Phat/rhohat)
     
     write(11,100) zc*scale_l, rc*scale_l,&
                   Phat*rhoinf*uinf**2,&
                   rhohat*rhoinf,&
                   That*const*rhoinf*uinf**2/rhoinf,&
                   uhat*uinf, sspeed*uinf,&
                   (Phat*rhoinf*uinf**2)/(rhohat*rhoinf)**gam
     
     rhat = rhat + meshsize
  end do
  
  close(11)
  
  if(verbose) write(*,'(a)') ">> Generated file: shock.dat."  
  
! (2) Write out profiles for fluid variables along the obstacle
! =============================================================

  rstar = floor_star
  
  open(12,file="body.dat",status="replace",action="write")
  
! generate body for given bow shock geometry
! use floor value for rhat to avoid singularity

  rhat = floor_hat

  
  do
     sinbhat2 = sinb2(rhat)
     cosbhat2 = cosb2(rhat)
     kappahat = kappa(rhat)
     
     llim = rstar**(1+j)/(1.d0+real(j,kind=8))
     ulim =  rhat**(1+j)/(1.d0+real(j,kind=8))
     
     chihat = (gam-1.d0)/(gam+1.d0)+2.d0/((gam+1.d0)*M**2*sinbhat2)
     Phat   = 1.d0/(gam*M**2)+(1.d0-chihat)*sinbhat2
     rhohat = 1.d0/chihat
     uhat   = sqrt(cosbhat2)
     That   = Phat/rhohat
     
     integr1sol = romberg_trap(llim,ulim,integr1,tol,iter)
     
     P      = 1.d0/(gam*M**2)+(1.d0-chihat)*sinbhat2-kappahat/rhat**j*integr1sol
     Pb     = P     
     rhob   = ((gam-1.d0)/(gam+1.d0)+2.d0/((gam+1.d0)*M**2*sinb2(rstar)))**(-1)*&
              ((Pb*sinbhat2)/(Phat*sinb2(rstar)))**(1.d0/gam)
     ub     = sqrt(cosb2(rstar)+(2.d0/((gam-1.d0)*M**2)+sinb2(rstar))*&
              (1.d0-((Pb*sinbhat2)/(Phat*sinb2(rstar)))**((gam-1.d0)/gam)))    
     Tb     = Pb/rhob
     sspeed = sqrt(gam*Pb/rhob)
     
     integr2sol = romberg_trap(llim,ulim,integr2,tol,iter)
     
     select case (j)
        case(0)
           delta = integr2sol
        case(1)
           delta = rhat/sqrt(cosbhat2)*(1.d0-sqrt(1.d0-(2.d0*integr2sol*&
                   sqrt(cosbhat2))/rhat**2))
     end select
     
!    re-transform
     zc = z(rhat)+delta*sqrt(sinbhat2)
     rc = rhat   -delta*sqrt(cosbhat2)
     
     if(zc > zhatmax) exit
     
     
     
     write(12,100) zc*scale_l, rc*scale_l,&
                   Pb*rhoinf*uinf**2,&
                   rhob*rhoinf,&
                   Tb*const*rhoinf*uinf**2/rhoinf,&
                   ub*uinf, sspeed*uinf, &
                   (Pb*rhoinf*uinf**2)/(rhob*rhoinf)**gam
    
     rhat = rhat + meshsize
  end do
  
  close(12)
  
  if(verbose) write(*,'(a)') ">> Generated file: body.dat."
  
  
! (3) Write out profiles for fluid variables along shock-layer probe lines
! ========================================================================

  do i = 1, plines
     zhat = zhatmax - zhatstep(i)*(zhatmax-zhatmin)
     rhat = r(zhat)
     write(out,"(i5.5)") i
     open(13,file="probe"//out//".dat",status="replace",action="write")
 
     rstar = floor_star
  
     do
        sinbhat2 = sinb2(rhat)
        cosbhat2 = cosb2(rhat)
        kappahat = kappa(rhat)
  
        llim = rstar**(1+j)/(1.d0+real(j,kind=8))
        ulim =  rhat**(1+j)/(1.d0+real(j,kind=8))
  
        if(llim > ulim) exit
    
        chihat = (gam-1.d0)/(gam+1.d0)+2.d0/((gam+1.d0)*M**2*sinbhat2)
        Phat   = 1.d0/(gam*M**2)+(1.d0-chihat)*sinbhat2
        rhohat = 1.d0/chihat       
        uhat   = sqrt(cosbhat2)
        That   = Phat/rhohat
        
        integr1sol = romberg_trap(llim,ulim,integr1,tol,iter)
        
        P      = 1.d0/(gam*M**2)+(1.d0-chihat)*sinbhat2-kappahat/rhat**j*integr1sol
        rho    = ((gam-1.d0)/(gam+1.d0)+2.d0/((gam+1.d0)*M**2*sinb2(rstar)))**(-1)*&
                 ((P*sinbhat2)/(Phat*sinb2(rstar)))**(1.d0/gam)
        u      = sqrt(cosb2(rstar)+(2.d0/((gam-1.d0)*M**2)+sinb2(rstar))*&
                 (1.d0-((P*sinbhat2)/(Phat*sinb2(rstar)))**((gam-1.d0)/gam)))
        T      = P/rho
        sspeed = sqrt(gam*P/rho)
        
        integr2sol = romberg_trap(llim,ulim,integr2,tol,iter)
        
        select case (j)
           case (0)
              y = integr2sol
           case (1)
              y = rhat/sqrt(cosbhat2)*(1.d0-sqrt(1.d0-(2.d0*integr2sol*&
                  sqrt(cosbhat2))/rhat**2))
        end select
        
        
!       re-transform
        zc = zhat +y*sqrt(sinbhat2)
        rc = rhat -y*sqrt(cosbhat2)
        

        write(13,101) zc*scale_l, rc*scale_l,&
                      P*rhoinf*uinf**2,&
                      rho*rhoinf,&
                      T*const*rhoinf*uinf**2/rhoinf,&
                      u*uinf, sspeed*uinf,&
                      (P*rhoinf*uinf**2)/(rho*rhoinf)**(gam),&
                      y*scale_l
        
        rstar = rstar + meshsize
     end do

     if(verbose) write(*,200,advance="no") achar(13),">> Generated ",i, " of ",plines,&
                 " add'l files; just finished: probe",i,".dat."

     close(13)
  end do

  deallocate(zhatstep)
  
  call system_clock (t2,clock_rate,clock_max )
  if(verbose) write(*,300) ">> Calculations completed after ",&
              real(t2-t1,kind=8)/real(clock_rate,kind=8)," seconds."
  
100 format(8(1x,E12.6))
101 format(9(1x,E12.6))
200 format(A1,A,t15,I5,A,I5,A,I5.5,A)
300 format(/A,F9.2,A)

  contains
         
         
     ! Bow shock equation
     real(kind=8) function r(z)
     real(kind=8),intent(in) :: z
        r = (a*z)**(1.d0/alpha)
     end function r
         
     ! Inverse bow shock equation
     real(kind=8) function z(r)
     real(kind=8),intent(in) :: r
        z = r**alpha/a
     end function z
         
     ! Analytical derivations
     ! q(z)=dr(z)/dz ---> q(r)
     real(kind=8) function q(r)
     real(kind=8),intent(in) :: r
        q = a/alpha*r**(1.d0-alpha)
     end function q
         
     ! f(z)=dq(z)/dz ---> f(r)
     real(kind=8) function f(r)
     real(kind=8),intent(in) :: r
        f = a**2/alpha*(1.d0/alpha-1.d0)*r**(1.d0-2.d0*alpha)
     end function f
         
         
     ! Additional functions
     real(kind=8) function kappa(r)
     real(kind=8),intent(in) :: r
        kappa = abs(f(r))/(1.d0+q(r)**2)**1.5d0
     end function kappa
         
     real(kind=8) function sinb2(r)
     real(kind=8),intent(in) :: r
        sinb2 = q(r)**2/(1.d0+q(r)**2)
     end function sinb2
         
     real(kind=8) function cosb2(r)
     real(kind=8),intent(in) :: r
        cosb2 = 1.d0/(1.d0+q(r)**2)
     end function cosb2
         
     real(kind=8) function sinbstar2(psi)
     real(kind=8),intent(in) :: psi
        sinbstar2 = 1.d0/((1.d0+real(j,kind=8))*psi)**(2.d0/(1.d0+real(j,kind=8)))/&
                    (1.d0+1.d0/((1.d0+real(j,kind=8))*psi)**(2.d0/(1.d0+real(j,kind=8))))
     end function sinbstar2
         
     real(kind=8) function cosbstar2(psi)
     real(kind=8),intent(in) :: psi
        cosbstar2 = 1.d0/(1.d0+1.d0/((1.d0+real(j,kind=8))*psi)**(2.d0/(1.d0+real(j,kind=8))))
     end function cosbstar2
         
     real(kind=8) function chistar(psi)
     real(kind=8),intent(in) :: psi
        chistar = (gam-1.d0)/(gam+1.d0)+2.d0/((gam+1.d0)*M**2*sinbstar2(psi))
     end function chistar
         
     real(kind=8) function integr1(psi)
     real(kind=8),intent(in) :: psi
        integr1 = sqrt(cosbstar2(psi)+(2.d0/((gam-1.d0)*M**2)+sinbstar2(psi))*&
                  (1.d0-(sinbhat2/sinbstar2(psi))**((gam-1.d0)/gam)))
     end function integr1        
      
     real(kind=8) function integr2(psi)
     real(kind=8),intent(in) :: psi        
        integr2 = chistar(psi)*(Phat*sinbstar2(psi)/(P*sinbhat2))**(1.d0/gam)/&
                  sqrt(cosbstar2(psi)+(2.d0/((gam-1.d0)*M**2)+sinbstar2(psi))*&
                  (1.d0-(P*sinbhat2/(Phat*sinbstar2(psi)))**((gam-1.d0)/gam)))
     end function integr2
      
  end program bluntbody
